import socket
from abc import ABC
from typing import List

import numpy as np

from .link import Link
from .matrix import Matrix


class Primitive(ABC):
    def __init__(self) -> None:
        self._json = ""

    def __str__(self) -> str:
        return self._json

    @staticmethod
    def _json_xyz(xyz: np.ndarray) -> str:
        return f'{{"x":{xyz[0]:.6},"y":{xyz[1]:.6},"z":{xyz[2]:.6}}}'

    @staticmethod
    def _json_quaternion(quaternion: np.ndarray) -> str:
        return f'{{"w":{quaternion[0]:.6},"x":{quaternion[1]:.6},"y":{quaternion[2]:.6},"z":{quaternion[3]:.6}}}'


class Line(Primitive):
    def __init__(self, start: np.ndarray, end: np.ndarray) -> None:
        super().__init__()

        self._json = f'{{"type":"line","start":{Primitive._json_xyz(start)},"end":{Primitive._json_xyz(end)}}}'


class Circle(Primitive):
    def __init__(self, xyz: np.ndarray, axis: np.ndarray, radius: float) -> None:
        super().__init__()

        self._json = f'{{"type":"circle","xyz":{Primitive._json_xyz(xyz)},"axis":{Primitive._json_xyz(axis)},"radius":{radius}}}'


class Dot(Primitive):
    def __init__(self, xyz: np.ndarray, size: int = 1) -> None:
        super().__init__()

        self._json = f'{{"type":"dot","xyz":{Primitive._json_xyz(xyz)},"size":{size}}}'


class Axes(Primitive):
    def __init__(self, matrix: Matrix, scale: int = 1) -> None:
        super().__init__()

        xyz = matrix.xyz
        quaternion = matrix.quaternion

        self._json = f'{{"type":"axes","xyz":{Primitive._json_xyz(xyz)},"quaternion":{Primitive._json_quaternion(quaternion)},"scale":{scale}}}'


class Label(Primitive):
    def __init__(self, xyz: np.ndarray, text: str) -> None:
        super().__init__()

        self._json = f'{{"type":"label","xyz":{Primitive._json_xyz(xyz)},"text":"{text}"}}'


def link_to_primitives(root: Link) -> List[Primitive]:
    primitives = []

    for link in root.flatten():
        joint = link.get_joint_global()
        end = link.get_end_global()

        primitives.append(Line(joint.xyz, end.xyz))
        primitives.append(Dot(joint.xyz))
        primitives.append(Axes(joint, 0.5 * link.length))

        imu = link.get_imu_global()

        primitives.append(Dot(imu.xyz, 0.5))
        primitives.append(Axes(imu, 0.25 * link.length))
        primitives.append(Label(imu.xyz, link.name))

        for next_link, _ in link.links:
            next_joint = next_link.get_joint_global()

            primitives.append(Line(joint.xyz, next_joint.xyz))
            primitives.append(Line(end.xyz, next_joint.xyz))

        wheel_axis = link.get_wheel_axis_global()

        if any(wheel_axis.xyz != 0):
            primitives.append(Circle(joint.xyz, wheel_axis.xyz, link.length))

    return primitives


class Connection:
    def __init__(self, ip_address: str = "localhost", port: int = 6000) -> None:
        self.__address = (ip_address, port)

        self.__socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        self.__socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 65535)

        self.__buffer_size = self.__socket.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF)

    def __del__(self) -> None:
        self.__socket.close()

    def send(self, primitives: List[Primitive]) -> None:
        json = "[" + ",".join([str(p) for p in primitives]) + "]"

        data = json.encode("ascii")

        if len(data) > self.__buffer_size:
            raise ValueError(f"The data size is {len(data)}, which exceeds the buffer size of {self.__buffer_size}.")

        self.__socket.sendto(data, self.__address)



import numpy as np

from ..link import Link
from ..matrix import Matrix

# Sets the height of the root so that the lowest part of the model is in
# contact with the floor.


def floor(root: Link) -> None:
    links = root.flatten()

    joints_z = np.array([l.get_joint_global().z for l in links])

    ends_z = np.array([l.get_end_global().z for l in links])

    wheels_z = np.array([_wheel_lowest_point(l)[2] for l in links if any(l.get_wheel_axis_global().xyz != 0)])

    min_z = np.min(np.concatenate((joints_z, ends_z, wheels_z)))

    root.joint = Matrix(x=root.joint.x, y=root.joint.y, z=root.joint.z - min_z, rotation=root.joint.rotation)


def _wheel_lowest_point(link: Link) -> np.ndarray:
    axis = link.get_wheel_axis_global().xyz

    lowest_spoke = np.cross(axis, np.cross(axis, np.array([0, 0, 1])))

    lowest_spoke /= np.linalg.norm(lowest_spoke)

    return link.get_joint_global().xyz + (link.length * lowest_spoke)



from typing import Tuple, Union

import numpy as np


class Matrix:
    def __init__(
        self,
        matrix: Union[None, np.matrix] = None,
        xyz: Union[None, np.ndarray] = None,
        x: float = 0,
        y: float = 0,
        z: float = 0,
        rotation: Union[None, np.ndarray] = None,
        quaternion: Union[None, np.ndarray] = None,
        axis_angle: Union[None, Tuple[np.ndarray, float]] = None,
        rot_xyz: Union[None, np.ndarray] = None,
        rot_x: float = 0,
        rot_y: float = 0,
        rot_z: float = 0,
    ) -> None:
        if xyz is not None:
            x, y, z = tuple(xyz)

        if rot_xyz is not None:
            rot_x, rot_y, rot_z = tuple(rot_xyz)

        if matrix is not None:
            self.__matrix = matrix.copy()
        elif rotation is not None:
            self.__matrix = np.matrix(
                [
                    [rotation[0, 0], rotation[0, 1], rotation[0, 2], x],
                    [rotation[1, 0], rotation[1, 1], rotation[1, 2], y],
                    [rotation[2, 0], rotation[2, 1], rotation[2, 2], z],
                    [0, 0, 0, 1],
                ]
            )
        elif quaternion is not None:
            # Quaternions and Rotation Sequence by Jack B. Kuipers, ISBN 0-691-10298-8, Page 168

            qw = quaternion[0]
            qx = quaternion[1]
            qy = quaternion[2]
            qz = quaternion[3]

            self.__matrix = np.matrix(
                [
                    [2 * (qw * qw - 0.5 + qx * qx), 2 * (qx * qy - qw * qz), 2 * (qx * qz + qw * qy), x],
                    [2 * (qx * qy + qw * qz), 2 * (qw * qw - 0.5 + qy * qy), 2 * (qy * qz - qw * qx), y],
                    [2 * (qx * qz - qw * qy), 2 * (qy * qz + qw * qx), 2 * (qw * qw - 0.5 + qz * qz), z],
                    [0, 0, 0, 1],
                ]
            )
        elif axis_angle is not None:
            # https://www.euclideanspace.com/maths/geometry/rotations/conversions/angleToQuaternion/index.htm

            axis, angle = axis_angle

            axis /= np.linalg.norm(axis)
            half_angle = np.radians(angle) / 2

            self.__matrix = Matrix(
                quaternion=[
                    np.cos(half_angle),
                    axis[0] * np.sin(half_angle),
                    axis[1] * np.sin(half_angle),
                    axis[2] * np.sin(half_angle),
                ]
            ).__matrix
        else:
            # rot_x, rot_y, rot_z are ZYX Euler angles
            # Quaternions and Rotation Sequence by Jack B. Kuipers, ISBN 0-691-10298-8, Page 167

            sx = np.sin(np.radians(rot_x))
            cx = np.cos(np.radians(rot_x))

            sy = np.sin(np.radians(rot_y))
            cy = np.cos(np.radians(rot_y))

            sz = np.sin(np.radians(rot_z))
            cz = np.cos(np.radians(rot_z))

            self.__matrix = np.matrix(
                [
                    [cz * cy, cz * sy * sx - sz * cx, cz * sy * cx + sz * sx, x],
                    [sz * cy, sz * sy * sx + cz * cx, sz * sy * cx - cz * sx, y],
                    [-sy, cy * sx, cy * cx, z],
                    [0, 0, 0, 1],
                ]
            )

    @property
    def xyz(self) -> np.ndarray:
        return np.array([self.__matrix[0, 3], self.__matrix[1, 3], self.__matrix[2, 3]])

    @property
    def x(self) -> float:
        return self.__matrix[0, 3]

    @property
    def y(self) -> float:
        return self.__matrix[1, 3]

    @property
    def z(self) -> float:
        return self.__matrix[2, 3]

    @property
    def rotation(self) -> np.matrix:
        return self.__matrix[0:3, 0:3]

    @property
    def quaternion(self) -> np.ndarray:
        # http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/

        m00 = self.__matrix[0, 0]
        m01 = self.__matrix[0, 1]
        m02 = self.__matrix[0, 2]
        m10 = self.__matrix[1, 0]
        m11 = self.__matrix[1, 1]
        m12 = self.__matrix[1, 2]
        m20 = self.__matrix[2, 0]
        m21 = self.__matrix[2, 1]
        m22 = self.__matrix[2, 2]

        tr = m00 + m11 + m22

        if tr > 0:
            s = np.sqrt(tr + 1) * 2  # s=4*w
            return np.array(
                [
                    0.25 * s,
                    (m21 - m12) / s,
                    (m02 - m20) / s,
                    (m10 - m01) / s,
                ]
            )
        elif m00 > m11 and m00 > m22:
            s = np.sqrt(1 + m00 - m11 - m22) * 2  # s=4*x
            return np.array(
                [
                    (m21 - m12) / s,
                    0.25 * s,
                    (m01 + m10) / s,
                    (m02 + m20) / s,
                ]
            )
        elif m11 > m22:
            s = np.sqrt(1 + m11 - m00 - m22) * 2  # s=4*y
            return np.array(
                [
                    (m02 - m20) / s,
                    (m01 + m10) / s,
                    0.25 * s,
                    (m12 + m21) / s,
                ]
            )
        else:
            s = np.sqrt(1 + m22 - m00 - m11) * 2  # s=4*z
            return np.array(
                [
                    (m10 - m01) / s,
                    (m02 + m20) / s,
                    (m12 + m21) / s,
                    0.25 * s,
                ]
            )

    @property
    def T(self) -> "Matrix":
        return Matrix(matrix=self.__matrix.T)

    def copy(self) -> "Matrix":
        return Matrix(matrix=self.__matrix)

    def __mul__(self, other: "Matrix") -> "Matrix":
        return Matrix(matrix=self.__matrix * other.__matrix)

    def __str__(self) -> str:
        return str(self.__matrix)



from ..link import Link
from ..matrix import Matrix

# Calibrates the alignment of all IMUs by comparing a predefined calibration
# pose with the IMU measurements while the subject recreates the pose.


def calibrate(
    root: Link,
    imus: dict[str, Matrix],  # {<link name>: <IMU measurment>, ...}
    calibraiton_pose: dict[str, Matrix] = {},  # {<link name>: <joint matrix>, ...}
) -> None:
    links = {l.name: l for l in root.flatten()}

    for link in links.values():
        link.joint = Matrix()

    for name, matrix in calibraiton_pose.items():
        links[name].joint = matrix

    for name, matrix in imus.items():
        links[name].set_imu_global(matrix)



from typing import Dict, List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import animation

from .matrix import Matrix


class Link:
    def __init__(self, name, end: Matrix, wheel_axis: Matrix = Matrix()) -> None:
        self.__name = str(name)
        self.__origin = Matrix()  # link origin in global frame
        self.__joint = Matrix()  # joint rotation relative to origin
        self.__end = end  # link end relative to origin
        self.__imu = Matrix(x=end.x / 2, y=end.y / 2, z=end.z / 2)  # IMU relative to origin
        self.__wheel_axis = wheel_axis  # direction defined by wheel_axis.xyz
        self.__links = []  # [(link, matrix), ...] where matrix is the origin of the next link relative to the end of this link
        self.__is_root = True

        if not np.isclose(np.dot(self.__end.xyz, self.__wheel_axis.xyz), 0):
            raise ValueError(f"Invalid values for {name}. Link end {self.__end.xyz} must be orthogonal to wheel_axis {self.__wheel_axis.xyz}.")

    @property
    def name(self) -> str:
        return self.__name

    @property
    def joint(self) -> Matrix:
        return self.__joint.copy()

    @joint.setter
    def joint(self, joint: Matrix) -> None:
        if self.__is_root:
            self.__joint = joint
        else:
            self.__joint = Matrix(x=self.__joint.x, y=self.__joint.y, z=self.__joint.z, rotation=joint.rotation)  # ignore joint.xyz
        self.__update()

    @property
    def imu(self) -> Matrix:
        return self.__imu.copy()

    @imu.setter
    def imu(self, imu: Matrix) -> None:
        self.__imu = Matrix(x=self.__imu.x, y=self.__imu.y, z=self.__imu.z, rotation=imu.rotation)  # ignore imu.xyz

    @property
    def links(self) -> List[Tuple["Link", Matrix]]:
        return self.__links

    @property
    def length(self) -> float:
        return np.linalg.norm(self.__end.xyz)

    def __update(self, origin: Union[None, Matrix] = None) -> None:
        if origin is not None:
            self.__origin = origin

        for link, matrix in self.__links:
            link.__update(self.__origin * self.joint * self.__end * matrix)

    def connect(self, link: "Link", matrix: Matrix = Matrix()) -> "Link":  # matrix is the origin of the next link relative to the end of this link
        link.__is_root = False
        self.__links.append((link, matrix))
        self.__update()
        return self

    def get_joint_global(self) -> Matrix:
        return self.__origin * self.__joint

    def get_end_global(self) -> Matrix:
        return self.__origin * self.__joint * self.__end

    def get_imu_global(self) -> Matrix:
        return self.__origin * self.__joint * self.__imu

    def set_imu_global(self, imu_global: Matrix) -> None:
        self.imu = self.__joint.T * self.__origin.T * imu_global  # transpose can be used instead of inverse because imu.xyz ignored

    def set_joint_from_imu_global(self, imu_global: Matrix) -> None:
        self.joint = self.__origin.T * imu_global * self.__imu.T  # transpose can be used instead of inverse because joint.xyz ignored

    def get_wheel_axis_global(self) -> Matrix:
        return Matrix(rotation=(self.__origin * self.__joint).rotation) * self.__wheel_axis

    def flatten(self) -> List["Link"]:
        links = [self]

        for link, _ in self.__links:
            links = np.concatenate((links, link.flatten()))

        return links

    def dictionary(self) -> Dict[str, "Link"]:
        return {l.name: l for l in self.flatten()}

    def plot(
        self,
        frames: Union[None, Dict[str, Matrix]] = None,  # each frame is a dictionary of joint matrices created by {l.name: l.joint for l in root.flatten()}
        fps: float = 30,  # animation frames per second
        file_name: str = "",  # must be .gif
        elev: Union[None, float] = None,  # see mpl_toolkits.mplot3d.axes3d.Axes3D.view_init
        azim: Union[None, float] = None,  # see mpl_toolkits.mplot3d.axes3d.Axes3D.view_init
        figsize: Union[None, Tuple[float, float]] = None,  # see matplotlib.pyplot.figure
        dpi: Union[None, float] = None,  # see matplotlib.pyplot.figure
        hide_tick_labels: bool = True,  # hides the tick labels
        block: Union[None, bool] = None,  # see matplotlib.pyplot.show
    ) -> None:
        links = self.flatten()

        # Create figure
        figure = plt.figure(figsize=figsize, dpi=dpi)

        axes = plt.axes(projection="3d")

        plt.subplots_adjust(top=0.95, bottom=0, left=0, right=1)

        if hide_tick_labels:
            axes.set_xticklabels([])
            axes.set_yticklabels([])
            axes.set_zticklabels([])

        # Create quivers, markers, and labels
        link_quivers = axes.quiver([], [], [], [], [], [], color="tab:gray", zorder=-np.inf, label="Link")

        x_quivers = axes.quiver([], [], [], [], [], [], color="tab:red", label="X")
        y_quivers = axes.quiver([], [], [], [], [], [], color="tab:green", label="Y")
        z_quivers = axes.quiver([], [], [], [], [], [], color="tab:blue", label="Z")

        (joint_markers,) = axes.plot([], [], [], "ko", markersize=4, zorder=np.inf, label="Joint")

        (imu_markers,) = axes.plot([], [], [], "ko", markersize=2, zorder=np.inf, label="IMU")

        labels = [axes.text(*l.get_imu_global().xyz, l.name, zorder=np.inf) for l in links]

        # Create index text
        if frames is not None:
            index_text = plt.figtext(0.99, 0.01, "", horizontalalignment="right")

        # Show legend
        axes.legend(loc="upper left", frameon=0)

        # Set view
        axes.view_init(elev=elev, azim=azim)

        # Update plot
        def update(index: Union[None, int] = None) -> None:
            # Set joints
            if index is not None:
                for name, joint in frames[index].items():
                    self.dictionary()[name].joint = joint

            # Set link quivers
            joints = np.array([l.get_joint_global() for l in links])
            ends = np.array([l.get_end_global() for l in links])

            link_quiver_segments = [[tuple(j.xyz), tuple(e.xyz)] for j, e in zip(joints, ends)]

            for joint, end, next_links in zip(joints, ends, [l.links for l in links]):
                for next_joint in [n.get_joint_global() for n, _ in next_links]:
                    link_quiver_segments.append([tuple(joint.xyz), tuple(next_joint.xyz)])
                    link_quiver_segments.append([tuple(end.xyz), tuple(next_joint.xyz)])

            for joint, end, wheel_axis in zip(joints, ends, [l.get_wheel_axis_global() for l in links]):
                if any(wheel_axis.xyz != 0):
                    chords = [joint.xyz + (Matrix(axis_angle=(wheel_axis.xyz, a)) * Matrix(xyz=end.xyz - joint.xyz)).xyz for a in np.linspace(0, 360, 36)]

                    for start, end in zip(chords, chords[1:]):
                        link_quiver_segments.append([tuple(start), tuple(end)])

            link_quivers.set_segments(link_quiver_segments)

            # Set XYZ quivers
            lengths = [0.5 * l.length for l in links]

            x_quiver_segments = [[tuple(j.xyz), tuple(j.xyz + (l * j.rotation[:, 0].A1))] for j, l in zip(joints, lengths)]
            y_quiver_segments = [[tuple(j.xyz), tuple(j.xyz + (l * j.rotation[:, 1].A1))] for j, l in zip(joints, lengths)]
            z_quiver_segments = [[tuple(j.xyz), tuple(j.xyz + (l * j.rotation[:, 2].A1))] for j, l in zip(joints, lengths)]

            imus = np.array([l.get_imu_global() for l in links])
            lengths = [0.25 * l.length for l in links]

            x_quiver_segments += [[tuple(i.xyz), tuple(i.xyz + (l * i.rotation[:, 0].A1))] for i, l in zip(imus, lengths)]
            y_quiver_segments += [[tuple(i.xyz), tuple(i.xyz + (l * i.rotation[:, 1].A1))] for i, l in zip(imus, lengths)]
            z_quiver_segments += [[tuple(i.xyz), tuple(i.xyz + (l * i.rotation[:, 2].A1))] for i, l in zip(imus, lengths)]

            x_quivers.set_segments(x_quiver_segments)
            y_quivers.set_segments(y_quiver_segments)
            z_quivers.set_segments(z_quiver_segments)

            # Set markers
            joint_markers.set_data([[j.xyz[0] for j in joints], [j.xyz[1] for j in joints]])
            joint_markers.set_3d_properties([j.xyz[2] for j in joints])

            imu_markers.set_data([[i.xyz[0] for i in imus], [i.xyz[1] for i in imus]])
            imu_markers.set_3d_properties([i.xyz[2] for i in imus])

            # Set labels text positions
            for label, imu in zip(labels, imus):
                label.set_position((imu.xyz[0], imu.xyz[1]))
                label.set_3d_properties(z=imu.xyz[2], zdir=None)

            # Set index text
            if frames is not None:
                index_text.set_text(f"Frame {index} of {len(frames)}")

            # Set limits
            all_xyz = np.concatenate(
                (
                    [l[1] for l in link_quiver_segments],
                    [x[1] for x in x_quiver_segments],
                    [y[1] for y in y_quiver_segments],
                    [z[1] for z in z_quiver_segments],
                )
            )

            axes.set_xlim3d(np.min(all_xyz[:, 0]), np.max(all_xyz[:, 0]))
            axes.set_ylim3d(np.min(all_xyz[:, 1]), np.max(all_xyz[:, 1]))
            axes.set_zlim3d(np.min(all_xyz[:, 2]), np.max(all_xyz[:, 2]))

            axes.set_box_aspect(np.ptp(all_xyz, axis=0))

        # Static plot
        if frames is None:
            update()
            plt.show(block=block)
            return

        # Animation
        anim = animation.FuncAnimation(figure, update, frames=len(frames), interval=1000 / fps, repeat=False, blit=False)

        if file_name:
            anim.save(file_name, writer=animation.PillowWriter(fps), dpi="figure", progress_callback=lambda i, n: print(f"Saving frame {i + 1} of {n}"))
        else:
            plt.show(block=block)  # play animation
